# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import moksha.common.utils
import copy
import time
import urllib
import uuid
import re
import os
import logging
import warnings
import json

import ConfigParser

from UserDict import DictMixin
from pytz import timezone, utc
from decorator import decorator
from kitchen.text.converters import to_unicode as kitchen_unicode

from moksha.common.exc import MokshaConfigNotFound
from moksha.common.lib.dates import DateTimeDisplay, utc_offset

log = logging.getLogger(__name__)
scrub_filter = re.compile('[^_a-zA-Z0-9-]')


@decorator
def trace(f, *args, **kw):
    """ A useful decorator for debugging method parameters and return values
    """
    r = None
    try:
        r = f(*args, **kw)
    finally:
        print "%s(%s, %s) = %s" % (f.func_name, args, kw, r)
    return r


# TODO -- remove this and just get it from pypi
try:
    from collections import defaultdict
except:
    # A pure Python version of Python 2.5's defaultdict
    # http://code.activestate.com/recipes/523034/
    # by Jason Kirtland
    class defaultdict(dict):
        def __init__(self, default_factory=None, *a, **kw):
            if (default_factory is not None and
                not hasattr(default_factory, '__call__')):
                raise TypeError('first argument must be callable')
            dict.__init__(self, *a, **kw)
            self.default_factory = default_factory

        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                return self.__missing__(key)

        def __missing__(self, key):
            if self.default_factory is None:
                raise KeyError(key)
            self[key] = value = self.default_factory()
            return value

        def __reduce__(self):
            if self.default_factory is None:
                args = tuple()
            else:
                args = self.default_factory,
            return type(self), args, None, None, self.items()

        def copy(self):
            return self.__copy__()

        def __copy__(self):
            return type(self)(self.default_factory, self)

        def __deepcopy__(self, memo):
            import copy
            return type(self)(self.default_factory,
                              copy.deepcopy(self.items()))

        def __repr__(self):
            return 'defaultdict(%s, %s)' % (self.default_factory,
                                            dict.__repr__(self))


def cache_rendered_data(data):
    """ A method to cache ``data`` with the current request path as the key.

    This method can be used within TurboGears2 hooks to cache rendered data
    from a given method for a specific URL.  For example, to cache the
    index.html, you could do something like this.

    .. code-block:: python

        from tg.decorators import after_render

        @after_render(cache_rendered_data)
        @expose('mako:moksha.templates.index')
        def index(self):
            return dict()

    :Warning: In this example usage, the method caches the data before it makes
              its way out of the WSGI middleware stack.  Therefore, widget
              resources are not injected, and stored in the cache.

    :Warning: This function only works with pylons.

    """
    import pylons

    if hasattr(pylons.g, 'cache') and pylons.g.cache and \
            pylons.request.environ.get('HTTP_X_FORWARDED_PROTO'):
        pylons.g.cache.set(pylons.request.path_qs, str(data))


@decorator
def cache_rendered(func, *args, **kwargs):
    content = func(*args, **kwargs)
    cache_rendered_data(content)
    return content


def in_full_moksha_stack():
    """
    Figure out if we are running Moksha as WSGI middleware, or in our full
    stack.

    :returns: True if we are currently running in Moksha's full WSGI stack,
              False if we are running Moksha only as WSGI middleware.
    """
    # This is deprecated, since there's no such thing as moksha 'full stack'
    # anymore.
    return False


def get_moksha_config_path():
    """
    :returns: The path to Moksha's configuration file.
    """
    for config_path in ('.', '/etc/moksha/', __file__ + '/../../../'):
        for config_file in ('production.ini', 'development.ini'):
            cfg = os.path.join(os.path.abspath(config_path), config_file)
            if os.path.isfile(cfg):
                return cfg

    log.warn('No moksha configuration file found, make sure the '
             'controlling app is fully configured')

    return None
    # raise MokshaConfigNotFound('Cannot find moksha configuration file!')


def get_moksha_dev_config():
    fname = 'development.ini'
    cfgs = [
        os.path.join(os.path.abspath(__file__ + '/../../../'), fname),
        os.path.join(os.path.abspath(__file__ + '/../../../../'), fname),
        os.path.join(os.getcwd(), fname),
        '/etc/moksha/%s' % fname,
    ]
    for cfg in cfgs:
        if os.path.isfile(cfg):
            return cfg
    log.warn("Cannot find configuration in %r" % cfgs)


def get_moksha_appconfig():
    """ Return the appconfig of Moksha """
    return appconfig('config:' + get_moksha_config_path())


def appconfig(config_path):
    """ Our own reimplementation of paste.deploy.appconfig """

    if config_path.startswith('config:'):
        config_path = config_path[7:]

    here = os.path.abspath(os.path.dirname(config_path))
    parser = ConfigParser.ConfigParser({"here": here})
    parser.read(filenames=[config_path])
    try:
        return dict(parser.items('app:main'))
    except ConfigParser.NoSectionError:
        for section in parser.sections():
            if section.startswith('app:'):
                print "Using %r" % section
                return dict(parser.items(section))

        raise ConfigParser.NoSectionError("Couldn't find app: section.")


def create_app_engine(app, config):
    """ Create a new SQLAlchemy engine for a given app """
    from sqlalchemy import create_engine
    return create_engine(config.get('app_db', 'sqlite:///%s.db') % app)


def to_unicode(obj, encoding='utf-8', errors='replace'):
    """
    :deprecated: by to_unicode in http://python-kitchen.fedorahosted.org
    """
    deprecation("to_unicode is deprecated in favor of "
                "kitchen.text.coverters.to_unicode().")
    return kitchen_unicode(obj, encoding, errors)


def replace_app_header(app, header_name, value):
        from paste.response import replace_header
        if app.headers:
            headers = list(app.headers)
        else:
            headers = []

        replace_header(headers, header_name, value)
        app.headers = headers


class EnumDataObj(dict):
    def __init__(self, code, data):
        super(EnumDataObj, self).__init__(code=code, data=data)

    def __getattribute__(self, name):
        try:
            return super(EnumDataObj, self).__getattribute__(name)
        except AttributeError, e:
            if name in self:
                return self[name]

            raise e

    def replace_app_header(self, app, header_name):
        replace_app_header(app, header_name, self.code)

    def __repr__(self):
        # act as if the user requested the code
        return str(self['code'])


class CategoryEnum(object):
    def __init__(self, prefix, *data):
        self._prefix = prefix
        self._code_map = {}

        for d in data:
            # data should be a tuple of (id, url)
            # id can not have any dots in them
            id = d[0]
            if id.find('.') != -1:
                raise ValueError(
                    'The enumeration id %s can not contain dots', id)

            # code is prefix.id
            code = '%s.%s' % (prefix, d[0])
            dob = EnumDataObj(code, d[1])
            setattr(self, d[0], dob)
            setattr(self, code, dob)
            self._code_map[code] = d[0]

    def is_valid_class(self, code):
        if code.beginswith(self._prefix + '_'):
            return True

        return False

    def code_to_attr(self, code):
        return self._code_map[code]

    def attr_to_code(self, attr):
        return self.__getattribute__(attr).code

    def attr_to_data(self, attr):
        return self.__getattribute__(attr).data

    def get_code(self, attr):
        return self.attr_to_code(attr)

    def get_data(self, attr):
        return self.attr_to_data(attr)

    def get_category(self):
        return self._prefix

    def __call__(self, code):
        return self.get_data(code)


class EnumGroup(object):
    def __init__(self):
        self._enums = {}

    def add(self, enum):
        self._enums[enum.get_category()] = enum

    def __getitem__(self, key):
        (category, enum_id) = key.rsplit('.', 1)
        enum = self._enums[category]
        return enum.get_data(key)


def strip_script(environ):
    """
    Strips the script portion of a url path so the middleware works even
    when mounted under a path other than root.
    """
    path = environ['PATH_INFO']
    if path.startswith('/') and 'SCRIPT_NAME' in environ:
        prefix = environ.get('SCRIPT_NAME')
        if prefix.endswith('/'):
            prefix = prefix[:-1]
        if path.startswith(prefix):
            path = path[len(prefix):]
    return path


def get_num_cpus():
    cpus = 1
    for line in open('/proc/cpuinfo'):
        if line.startswith('processor'):
            cpus = int(line.split()[-1]) + 1
    return cpus


def deprecation(message):
    warnings.warn(message, DeprecationWarning)


def listify(something):
    if something:
        return not isinstance(something, list) and [something] or something
    else:
        return []


class odict(DictMixin):

    def __init__(self):
        self._keys = []
        self._data = {}

    def index(self, i):
        k = self._keys[i]
        return self._data[k]

    def key_index(self, i):
        return self._keys[i]

    def __setitem__(self, key, value):
        if key not in self._data:
            self._keys.append(key)

        self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]
        self._keys.remove(key)

    def __iter__(self):
        for key in self._keys:
            yield key

    def keys(self):
        return list(self._keys)

    def copy(self):
        copyDict = odict()
        copyDict._data = self._data.copy()
        copyDict._keys = self._keys[:]
        return copyDict

    def __repr__(self):
        result = []
        for key in self._keys:
            result.append('(%s, %s)' % (repr(key), repr(self._data[key])))
        return ''.join(['OrderedDict', '([', ', '.join(result), '])'])



import pkg_resources
import re
import optparse
import os
import stat
import sys
import imp

from paste.script import command
from paste.script import create_distro

beginning_letter = re.compile(r"^[^a-z]*")
valid_only = re.compile(r"[^a-z0-9_]")

class MokshaQuickstartCommand(command.Command):
    """Create new Moksha components.

Create a new Moksha App with this command.

Example usage::

    $ paster moksha yourproject

    """
    version = pkg_resources.get_distribution('moksha.common').version
    max_args = 3
    min_args = 0
    summary = __doc__.splitlines()[0]
    usage = '\n' + __doc__
    group_name = "Moksha"
    name = None
    package = None
    svn_repository = None
    templates = "moksha.master"
    #dry_run = False
    #no_input = False

    livewidget = False
    consumer = False
    stream = False

    topic = 'moksha.test'

    parser = command.Command.standard_parser(quiet=True)
    parser = optparse.OptionParser(
                    usage="%prog moksha [options] [project name]",
                    version="%prog " + version)
    parser.add_option("-l", "--livewidget",
            help='Create an example Moksha LiveWidget',
            action="store_true", dest="livewidget")
    parser.add_option("-u", "--consumer",
            help='Create an example Moksha Consumer',
            action="store_true", dest="consumer")
    parser.add_option("-C", "--controller",
            help='Create an example Moksha Controller',
            action="store_true", dest="controller")
    parser.add_option("-P", "--producer",
            help='Create an example Moksha Producer',
            action="store_true", dest="stream")
    parser.add_option("-p", "--package",
            help="package name for the code",
            dest="package")
    parser.add_option("-t", "--topic",
            help="The Moksha topic to utilize",
            dest="topic")
    #parser.add_option("--dry-run",
    #        help="dry run (don't actually do anything)",
    #        action="store_true", dest="dry_run")
    #parser.add_option("--noinput",
    #        help="no input (don't ask any questions)",
    #        action="store_true", dest="no_input")

    def command(self):
        """Quickstarts the new project."""

        self.__dict__.update(self.options.__dict__)

        if self.args:
            self.name = self.args[0]

        while not self.name:
            self.name = raw_input("Enter project name: ")

        package = self.name.lower()
        package = beginning_letter.sub("", package)
        package = valid_only.sub("", package)
        if package:
            self.package = package
        else:
            self.package = None
            while not self.package:
                self.package = raw_input(
                    "Enter package name [%s]: " % package).strip() or package

        self.name = pkg_resources.safe_name(self.name).replace('-', '_')
        self.rpm_name = self.package.replace('.', '-')

        env = pkg_resources.Environment()
        if self.name.lower() in env:
            print 'The name "%s" is already in use by' % self.name,
            for dist in env[self.name]:
                print dist
                return

        try:
            if imp.find_module(self.package):
                print 'The package name "%s" is already in use' % self.package
                return
        except ImportError:
            pass

        if os.path.exists(self.name):
            print 'A directory called "%s" already exists. Exiting.' % self.name
            return

        command = create_distro.CreateDistroCommand("create")
        cmd_args = ['--template=moksha.master']
        if self.livewidget:
            cmd_args.append('--template=moksha.livewidget')
        if self.stream:
            cmd_args.append('--template=moksha.stream')
        if self.consumer:
            cmd_args.append('--template=moksha.consumer')
        if self.controller:
            cmd_args.append('--template=moksha.controller')

        #if self.dry_run:
        #    cmd_args.append("--simulate")
        #    cmd_args.append("-q")
        cmd_args.append(self.name)
        cmd_args.append("livewidget=%s" % self.livewidget)
        cmd_args.append("consumer=%s" % self.consumer)
        cmd_args.append("controller=%s" % self.controller)
        cmd_args.append("stream=%s" % self.stream)
        cmd_args.append("package=%s" % self.package)
        cmd_args.append("widget_name=%s" % self.package.title() + 'Widget')
        cmd_args.append("stream_name=%s" % self.package.title() + 'Stream')
        cmd_args.append("consumer_name=%s" % self.package.title() + 'Consumer')
        cmd_args.append("controller_name=%s" % self.package.title() + 'Controller')
        command.run(cmd_args)

        #if not self.dry_run:
        #    os.chdir(self.name)

        #    startscript = "start-%s.py" % self.package
        #    if os.path.exists(startscript):
        #        oldmode = os.stat(startscript).st_mode
        #        os.chmod(startscript, oldmode | stat.S_IXUSR)

        #    os.system('paver egg_info')

        #    # dirty hack to allow "empty" dirs
        #    for base, path, files in os.walk("./"):
        #        for file in files:
        #            if file == "empty":
        #                os.remove(os.path.join(base, file))



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
log = logging.getLogger(__name__)

from moksha.exc import CacheBackendException

class Cache(object):
    """ A memcached-specific caching interface """

    def __init__(self, url, timeout=None, prefix=None):
        try:
            import memcache
        except ImportError:
            log.warning('Cannot import the `memcache` module.  Install the '
                        'python-memcached package to enable Mokshas memcached '
                        'integration.')
            return
        self.mc = memcache.Client([url])
        self.timeout = timeout or 300
        self.prefix = prefix
        if not self.mc.set('x', 'x', 1):
            raise CacheBackendException("Cannot connect to Memcached")

    def get(self, key):
        return self.mc.get(key)

    def set(self, key, value, timeout=None):
        if self.prefix:
            key = self.prefix + key
        self.mc.set(key, value, timeout or self.timeout)



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utilities to make writing tests easier """

import functools
import socket
import unittest

from moksha.common.lib.helpers import get_moksha_appconfig


def crosstest(method):
    @functools.wraps(method)
    def wrapper(self):
        for setup, name in self._setUp():
            def _inner(name):
                setup()
                try:
                    return method(self)
                finally:
                    self._tearDown()

            yield _inner, name

    return wrapper

config_sets = {
    'stomp': {
        "moksha.livesocket": "True",
        "moksha.livesocket.backend": "stomp",
        "orbited_host": "localhost",
        "orbited_port": "9000",
        "orbited_scheme": "http",
        "stomp_broker": "localhost",
        "stomp_port": "61613",
        "stomp_user": "guest",
        "stomp_pass": "guest",
    },
    'amqp': {
        "moksha.livesocket": "True",
        "moksha.livesocket.backend": "amqp",
        "orbited_host": "localhost",
        "orbited_port": "9000",
        "orbited_scheme": "http",
        "amqp_broker": "guest/guest@localhost",
        "amqp_broker_host": "localhost",
        "amqp_broker_port": "5672",
        "amqp_broker_user": "guest",
        "amqp_broker_pass": "guest",
        "amqp_broker_ssl": "False",
    },
    'zeromq': {
        "moksha.livesocket": "True",
        "moksha.livesocket.backend": "websocket",
        "moksha.livesocket.websocket.port": "9998",
        "zmq_enabled": "True",
        "zmq_publish_endpoints": "tcp://*:6543",
        "zmq_subscribe_endpoints": "tcp://127.0.0.1:6543",
        "zmq_strict": "True",
    },
}

flash_keys = []
for name, config_set in config_sets.items():
    flash_keys.extend(config_set.keys())

flash_keys = list(set(flash_keys))


def should_skip_config_set(name, config_set):
    if name == 'stomp':
        return True  # TODO - remove this.  This disables all STOMP tests.
        address = (config_set['stomp_broker'],
                   config_set['stomp_port'])
        # If we can connect, then run tests.  If we can't, then don't.
        try:
            s = socket.create_connection(address)
            s.close()
            return False
        except:
            return True
    elif name == 'amqp':
        address = (config_set['amqp_broker_host'],
                   config_set['amqp_broker_port'])

        # If we can't import qpid, then bail... but also:
        # If we can connect, then run tests.  If we can't, then don't.
        try:
            import qpid
            s = socket.create_connection(address)
            s.close()
            return False
        except:
            return True
    elif name == 'zeromq':
        return False
    else:
        raise ValueError("Unknown config set name.")


def make_setup_functions(kernel):
    for name, config_set in config_sets.items():

        def __setup():

            if should_skip_config_set(name, config_set):
                raise unittest.SkipTest("%s is not available." % name)

            config = get_moksha_appconfig()

            for key in flash_keys:
                if key in config:
                    del config[key]

            for key, value in config_set.items():
                config[key] = value

            kernel(config)

        yield __setup, name



import mock

# This is moved in python3.
# TODO -- use six.StringIO
import StringIO

from nose.tools import eq_

import moksha.common.lib.helpers

@mock.patch("sys.stdout", new_callable=StringIO.StringIO)
def test_trace(mock_stdout):

    @moksha.common.lib.helpers.trace
    def my_function(a, b):
        return None

    my_function(2, b=4)

    value = mock_stdout.getvalue()
    eq_(value, "my_function((2, 4), {}) = None\n")



""" The Moksha Command-line Interface """

import os
import sys
import signal
import logging
import pkg_resources

from optparse import OptionParser
from twisted.internet import protocol
from twisted.internet import reactor

log = logging.getLogger(__name__)

pids = []

class MokshaProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, name):
        self.name = name
    def connectionMade(self):
        pass
    def outReceived(self, data):
        sys.stdout.write(data)
    def errReceived(self, data):
        sys.stderr.write(data)
    def inConnectionLost(self):
        pass
    def outConnectionLost(self):
        pass
    def errConnectionLost(self):
        pass
    def processEnded(self, status_object):
        print "Process %r quit with status %d" % (
                self.name, status_object.value.exitCode)
        reactor.stop()
        for pid in pids:
            try:
                os.kill(pid, signal.SIGTERM)
            except OSError:
                pass


class MokshaCLI(object):

    def _exec(self, process, *args, **kw):
        args = args and [process] + list(args) or [process]
        print "Running %s" % ' '.join(args)
        pp = MokshaProcessProtocol(name=process)
        process = reactor.spawnProcess(pp, process, args,
                env={'PYTHONPATH': os.getcwd()}, **kw)
        pids.append(process.pid)

    def start(self):
        """ Start all of the Moksha components """

        from moksha.lib.helpers import get_moksha_config_path

        orbited = ['orbited']
        if os.path.exists('/etc/moksha/orbited.cfg'):
            orbited += ['-c', '/etc/moksha/orbited.cfg']

        self._exec(*orbited)
        self._exec('paster', 'serve', get_moksha_config_path())
        self._exec('moksha-hub', '-v')

    def list(self):
        """ List all available apps, widgets, producers and consumers """
        entry_points = ('root', 'widget', 'application', 'wsgiapp',
                        'producer', 'consumer')
        for entry in entry_points:
            print "[moksha.%s]" % entry
            for obj_entry in pkg_resources.iter_entry_points('moksha.' + entry):
                print " * %s" % obj_entry.name
            print

    def install(self):
        """ Install a Moksha component """

    def uninstall(self):
        """ Uninstall a Moksha component """

    def quickstart(self):
        """ Create a new Moksha component """
        # If no arguments given, run `paster moksha --help`

    def send(self, topic, message):
        """ Send a message to a topic """
        from moksha.api.hub import MokshaHub, reactor
        hub = MokshaHub()
        print "send_message(%s, %s)" % (topic, message)
        hub.send_message(topic, {'msg': message})

        def stop_reactor():
            hub.close()
            reactor.stop()

        reactor.callLater(0.2, stop_reactor)
        reactor.run()


def get_parser():
    usage = 'usage: %prog [command]'
    parser = OptionParser(usage, description=__doc__)
    parser.add_option('', '--start', action='store_true', dest='start',
                      help='Start Moksha')
    parser.add_option('', '--list', action='store_true', dest='list',
                      help='List all installed Moksha components')
    parser.add_option('', '--send', action='store_true', dest='send',
        help='Send a message to a given topic. Usage: send <topic> <message>')
    return parser


def main():
    parser = get_parser()
    opts, args = parser.parse_args()
    pkg_resources.working_set.add_entry(os.getcwd())

    moksha = MokshaCLI()

    logging.basicConfig(level=logging.INFO, format=
            '%(asctime)s,%(msecs)03d %(levelname)-5.5s [%(name)s] %(message)s')

    stdout = logging.StreamHandler(sys.stdout)
    stdout.setFormatter(logging.Formatter('%(message)s'))
    log.addHandler(stdout)

    if opts.start or 'start' in args:
        print "Starting Moksha..."
        moksha.start()
        try:
            reactor.run()
        except Exception, e:
            print "Caught exception: %s" % str(e)
            moksha.stop()
    elif opts.list or 'list' in args:
        moksha.list()
    elif opts.send or 'send' in args:
        if len(sys.argv) != 4:
            log.error('Usage: moksha send <topic> <message>')
            sys.exit(-1)
        moksha.send(sys.argv[2], sys.argv[3])
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import datetime


def utc_offset(tz):
    """ Return the UTC offset for a given timezone.

        >>> utc_offset('US/Eastern')
        '-4'

    """
    utc_offset = ''
    now = datetime.now(utc)
    now = now.astimezone(timezone(tz))
    offset = now.strftime('%z')
    if offset.startswith('-'):
        offset = offset[1:]
        utc_offset += '-'
    hours = int(offset[:2])
    utc_offset += str(hours)
    # FIXME: account for minutes?
    #minutes = int(offset[2:])
    #if minutes:
    #    utc_offset += '.%d' % ...
    return utc_offset


class DateTimeDisplay(object):
    """
    DateTimeDisplay is an object which takes any number of datetime objects
    and process them for display::

        >>> from datetime import datetime
        >>> now = datetime(2009, 5, 12)
        >>> later = datetime(2009, 5, 13)
        >>> d = DateTimeDisplay(now)
        >>> print d
        2009-05-12 00:00:00
        >>> d.age(later)
        '1 day'
        >>> d.age(datetime(2010, 7, 10, 10, 10), granularity='minute')
        '1 year, 1 month, 29 days, 10 hours and 10 minutes'
        >>> d.age(datetime(2010, 7, 10, 10, 10), tz='Europe/Amsterdam')
        '1 year, 1 month, 29 days and 10 hours'
        >>> d = DateTimeDisplay(datetime(2009, 5, 12, 12, 0, 0))
        >>> d.timestamp
        datetime.datetime(2009, 5, 12, 12, 0)
        >>> d.astimezone('Europe/Amsterdam')
        datetime.datetime(2009, 5, 12, 14, 0, tzinfo=<DstTzInfo 'Europe/Amsterdam' CEST+2:00:00 DST>)

    """
    def __init__(self, timestamp, format='%Y-%m-%d %H:%M:%S'):
        if isinstance(timestamp, basestring) and '.' in timestamp:
            timestamp = timestamp.split('.')[0]
        self.timestamp = timestamp
        if isinstance(timestamp, datetime.datetime):
            self.datetime = timestamp
        elif isinstance(timestamp, time.struct_time):
            self.datetime = datetime.datetime(*timestamp[:-2])
        elif isinstance(timestamp, basestring):
            if hasattr(datetime, 'strptime'): # Python 2.5+
                self.datetime = datetime.datetime.strptime(timestamp, format)
            else: # Python 2.4
                self.datetime = datetime.datetime(*time.strptime(timestamp, format)[:-2])
        else:
            raise Exception("You must provide either a datetime object or a"
                            "string, not %s" % type(timestamp))

    def astimezone(self, tz):
        """ Return `self.datetime` as a different timezone """
        timestamp = self.datetime.replace(tzinfo=utc)
        zone = timezone(tz)
        return zone.normalize(timestamp.astimezone(zone))

    def age(self, end=None, tz=None, granularity='hour', general=False):
        """
        Return the distance of time in words from `self.datetime` to `end`.

            >>> start = datetime(1984, 11, 02)
            >>> now = datetime(2009, 5, 22, 12, 11, 10)
            >>> DateTimeDisplay(start).age(now)
            '2 decades, 4 years, 6 months, 20 days and 12 hours'
            >>> DateTimeDisplay(start).age(now, general=True)
            '2 decades'

        """
        start = self.datetime
        if not end:
            end = datetime.datetime.utcnow()
        else:
            if isinstance(end, DateTimeDisplay):
                end = end.datetime
        if tz:
            zone = timezone(tz)
            end = end.replace(tzinfo=utc)
            end = zone.normalize(end.astimezone(zone))
            start = self.astimezone(tz)

        age = distance_of_time_in_words(start, end, granularity=granularity)

        if general:
            if not age.startswith('less than'):
                age = age.split('and')[0].split(',')[0]

        return age

    def __str__(self):
        return self.datetime.strftime('%Y-%m-%d %H:%M:%S %Z%z')

    def __repr__(self):
        return "<DateTimeDisplay %r>" % self.datetime

# The following functions were all copied *wholesale* from webhelpers.date

def _process_carryover(deltas, carry_over):
    """A helper function to process negative deltas based on the deltas
    and the list of tuples that contain the carry over values"""
    for smaller, larger, amount in carry_over:
        if deltas[smaller] < 0:
            deltas[larger] -= 1
            deltas[smaller] += amount


def _pluralize_granularity(granularity):
    """Pluralize the given granularity"""
    if 'century' == granularity:
        return "centuries"
    return granularity + "s"


def _delta_string(delta, granularity):
    """Return the string to use for the given delta and ordinality"""
    if 1 == delta:
        return "1 " + granularity
    elif delta > 1:
        return str(delta) + " " + _pluralize_granularity(granularity)


def _is_leap_year(year):
    if year % 4 == 0 and year % 400 != 0:
        return True
    return False


def distance_of_time_in_words(from_time, to_time=0, granularity="second",
                              round=False):
    """
    Return the absolute time-distance string for two datetime objects,
    ints or any combination you can dream of.

    If times are integers, they are interpreted as seconds from now.

    ``granularity`` dictates where the string calculation is stopped.
    If set to seconds (default) you will receive the full string. If
    another accuracy is supplied you will receive an approximation.
    Available granularities are:
    'century', 'decade', 'year', 'month', 'day', 'hour', 'minute',
    'second'

    Setting ``round`` to true will increase the result by 1 if the fractional
    value is greater than 50% of the granularity unit.

    Examples:

    >>> distance_of_time_in_words(86399, round=True, granularity='day')
    '1 day'
    >>> distance_of_time_in_words(86399, granularity='day')
    'less than 1 day'
    >>> distance_of_time_in_words(86399)
    '23 hours, 59 minutes and 59 seconds'
    >>> distance_of_time_in_words(datetime(2008,3,21, 16,34),
    ... datetime(2008,2,6,9,45))
    '1 month, 15 days, 6 hours and 49 minutes'
    >>> distance_of_time_in_words(datetime(2008,3,21, 16,34),
    ... datetime(2008,2,6,9,45), granularity='decade')
    'less than 1 decade'
    >>> distance_of_time_in_words(datetime(2008,3,21, 16,34),
    ... datetime(2008,2,6,9,45), granularity='second')
    '1 month, 15 days, 6 hours and 49 minutes'
    """
    granularities = ['century', 'decade', 'year', 'month', 'day', 'hour',
                     'minute', 'second']

    # 15 days in the month is a gross approximation, but this
    # value is only used if rounding to the nearest month
    granularity_size = {'century': 10, 'decade': 10, 'year': 10, 'month': 12,
                        'day': 15, 'hour': 24, 'minute': 60, 'second': 60 }

    if granularity not in granularities:
        raise ValueError("Please provide a valid granularity: %s" %
                        (granularities))

    # Get everything into datetimes
    if isinstance(from_time, int):
        from_time = datetime.datetime.fromtimestamp(time.time()+from_time)

    if isinstance(to_time, int):
        to_time = datetime.datetime.fromtimestamp(time.time()+to_time)

    # Ensure that the to_time is the larger
    if from_time > to_time:
        s = from_time
        from_time = to_time
        to_time = s
    # Stop if the tiems are equal
    elif from_time == to_time:
        return "0 " + _pluralize_granularity(granularity)

    # Collect up all the differences
    deltas = {'century': 0, 'decade': 0, 'year': 0, 'month': 0, 'day': 0,
              'hour': 0, 'minute': 0, 'second' : 0}

    # Collect the easy deltas
    for field in ['month', 'hour', 'day', 'minute', 'second']:
        deltas[field] = getattr(to_time,field) - getattr(from_time,field)

    # deal with year, century and decade
    delta_year = to_time.year - from_time.year
    if delta_year >= 100:
        deltas['century'] = delta_year // 100
    if delta_year % 100 >= 10:
        deltas['decade'] = delta_year // 10 - deltas['century'] * 10
    if delta_year % 10:
        deltas['year'] = delta_year % 10

    # Now we need to deal with the negative deltas, as we move from
    # the smallest granularity to the largest when we encounter a negative
    # we will 'borrow' from the next highest value.  Because to_time is
    # the larger of the two,
    carry_over = [('second', 'minute', granularity_size['second']),
                  ('minute', 'hour', granularity_size['minute']),
                  ('hour', 'day', granularity_size['hour'])]

    _process_carryover(deltas, carry_over)

    # Day is its own special animal.  We need to deal with negative days
    # differently depending on what months we are spanning across.  We need to
    # look up the from_time.month value in order to bring the number of days
    # to the end of the month.
    month_carry = [None, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if deltas['day'] < 0:
        deltas['month'] -= 1
        # Deal with leap years
        if (from_time.month) == 2 and _is_leap_year(from_time.year):
            deltas['day'] += 29
        else:
            deltas['day'] += month_carry[from_time.month]

    carry_over = [('month', 'year', granularity_size['month']),
                  ('year', 'decade', granularity_size['year']),
                  ('decade', 'century', granularity_size['decade'])]

    _process_carryover(deltas, carry_over)

    # Display the differences we care about, at this point we should only have
    # positive deltas
    return_strings = []
    for g in granularities:
        delta = deltas[g]
        # This is the finest granularity we will display
        if g == granularity:
            # We can only use rounding if the granularity is higher than
            # seconds
            if round and g != 'second':
                i = granularities.index(g)
                # Get the next finest granularity and it's delta
                g_p = granularities[i + 1]
                delta_p = deltas[g_p]
                # Determine if we should round up
                if delta_p > granularity_size[g_p] / 2:
                    delta += 1

                if delta != 0:
                    return_strings.append(_delta_string(delta, g))

                if not return_strings:
                    return "less than 1 " + granularity
                break

            else:
                if delta != 0:
                    return_strings.append(_delta_string(delta, g))

                # We're not rounding, check to see if we have encountered
                # any deltas to display, if not our time difference
                # is less than our finest granularity
                if not return_strings:
                    return "less than 1 " + granularity
                break
        # Read the value and continue
        else:
            if delta != 0:
                return_strings.append(_delta_string(delta, g))

    if len(return_strings) == 1:
        return return_strings[0]
    return ", ".join(return_strings[:-1]) + " and " + return_strings[-1]


def time_ago_in_words(from_time, granularity="second", round=False):
    """
    Return approximate-time-distance string for ``from_time`` till now.

    Same as ``distance_of_time_in_words`` but the endpoint is now.
    """
    return distance_of_time_in_words(from_time, datetime.datetime.now(),
        granularity, round)




# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

# This is required (oddly) to fix a python 2.7 bug with nose tests.
try:
    import multiprocessing, logging
except Exception:
    pass

setup(
    name='moksha.common',
    version='1.0.1',
    description='Common components for Moksha',
    author='Luke Macken, John (J5) Palmieri, Mairin Duffy, and Ralph Bean',
    author_email='',
    url='http://moksha.fedorahosted.org',
    install_requires=[
        "decorator",
        "paste",
        "pytz",
        "kitchen",
    ],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    namespace_packages=['moksha'],
    tests_require=[
        'nose',
        'mock',
    ],
    entry_points="""
    [console_scripts]
    moksha = moksha.common.commands.cli:main

    [paste.global_paster_command]
    moksha = moksha.common.commands.quickstart:MokshaQuickstartCommand

    """,
)



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__import__('pkg_resources').declare_namespace(__name__)

from paste.registry import StackedObjectProxy

# The root controller class
root = None

# The central feed cache, used by the Feed widget.
feed_cache = None

# A dict-like persistent backing store
feed_storage = None

# All loaded moksha applications
_apps = None

# All loaded ToscaWidgets
_widgets = None

# All WSGI applications
wsgiapps = None

# All loaded moksha menus
menus = None

# Per-request topic callbacks registered by rendered live widgets
livewidgets = StackedObjectProxy(name="livewidgets")

def get_widget(name):
    """ Get a widget instance by name """
    if _widgets:
        return _widgets[name]['widget']

def get_widgets():
    """ Return a dictionary of all widgets """
    from pkg_resources import iter_entry_points
    return _widgets or [widget.load() for widget in iter_entry_points('moksha.widget')]

def get_app(name):
    """ Get an app controller by name """
    return _apps[name]['controller']

def shutdown():
    """ Called when Moksha shuts down """
    try:
        if feed_storage:
            feed_storage.close()
    except AttributeError:
        pass
